/* SPDX-License-Identifier: BSD-2-Clause */
/*
 * Copyright (C) 2021 - 2023, Raspberry Pi Ltd
 *
 * version.hpp - libpisp auto-generated versioning
 */
#pragma once

#include <string>

namespace libpisp {

const std::string& version();

} // namespace libpisp
